package org.tekkotsu.ui.rcp;

import java.io.PrintStream;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Map;

import org.eclipse.equinox.app.IApplication;
import org.eclipse.equinox.app.IApplicationContext;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.PlatformUI;
import org.tekkotsu.ui.editor.StateMachineEditor;
import org.tekkotsu.ui.settings.SettingsFile;
import org.tekkotsu.ui.storyboard.views.StoryboardViewer;
import org.tekkotsu.ui.util.Debugger;
import org.tekkotsu.ui.util.FilesystemUtil;

public class StoryboardApplication implements IApplication {
	

	@SuppressWarnings("unchecked")
	@Override
	public Object start(IApplicationContext context) {
        Display display = PlatformUI.createDisplay();

        try {
    		SettingsFile.loadSettings();
    		Debugger.printDebug("Loaded: " + FilesystemUtil.stickyPathGet());
    		
    		Map<String, String[]> argmap = context.getArguments();
    		String[] args = argmap.get("application.args");
    		
    		parseCommandArguments(args);
        	
            int returnCode = PlatformUI.createAndRunWorkbench(display,
                    new StoryboardWorkbenchAdvisor());
            if (returnCode == PlatformUI.RETURN_RESTART) {
                return IApplication.EXIT_RESTART;
            }
            
            return IApplication.EXIT_OK;
        } catch( Exception e ) {
        	Debugger.force();
        	Debugger.printThrowable(e);
        	return IApplication.EXIT_OK;
        } finally {
    		SettingsFile.saveSettings();
    		FilesystemUtil.removeTempFiles();
    		Debugger.printDebug("Saved: " + FilesystemUtil.stickyPathGet());
        	try {
	            display.dispose();
        	} catch (Exception e) {
        		//nothing
        	}
        }
	}

	private void parseCommandArguments(String[] args) {
		for (int i=0;i<args.length;i++)
			Debugger.printDebug(Debugger.DEBUG_ALL, args[i]);
		
		PrintStream out = System.out;
		
		if( args.length > 0 && args[0].contains("help") )
		{
			out.println("Command Line Args");
			out.println();
			out.println("--host (-h) <hostname> :: Sets the hostname on startup");
			out.println("--behavior (-b) <\"name of behavior\"> :: Sets the behavior name on startup");
			out.println("--layout (-l) <path> :: Loads a layout file on startup");
			out.println("--trace (-t) <path> :: Loads a trace file on startup");
			out.println("--debug :: Equivalent to --debuglevel <max level>");
			out.println("--debuglevel <integer level> :: sets the debug level (higher is more debug, zero is no debug)");
			System.exit(0);
		}
		
		Iterator<String> argIter = Arrays.asList(args).iterator();
		while( argIter.hasNext() )
		{
			String s = argIter.next();
			if( s.equals("-h") || s.equals("--host") )
			{
				if( argIter.hasNext() )
				{
					Global.startHostName = argIter.next();
				}
			}
			else if( s.equals("-b") || s.equals("--behavior") )
			{
				if( argIter.hasNext() )
				{
					Global.startMachineName = argIter.next();
				}
			}
			else if( s.equals("-l") || s.equals("--layout") )
			{
				if( argIter.hasNext() )
				{
					String path = argIter.next();
					
					//if the path is relative
					if( !path.startsWith("/") )
					{
						path = FilesystemUtil.PWD + "/" + path;
					}
					
					StateMachineEditor.startupLayoutFile = path;
				}
			}
			else if( s.equals("-t") || s.equals("--trace") )
			{
				if( argIter.hasNext() )
				{
					String path = argIter.next();
					
					//if the path is relative
					if( !path.startsWith("/") )
					{
						path = FilesystemUtil.PWD + "/" + path;
					}
					
					StoryboardViewer.startupTraceFile = path;
				}
			}
			else if( s.equals("--debug") )
			{
				Debugger.debugLevel = Debugger.DEBUG_ALL;
			}
			else if (s.equals("--debuglevel"))
			{
				if (argIter.hasNext()) {
					String level = argIter.next();

					try {
						Debugger.debugLevel = Integer.parseInt(level);
					} catch (NumberFormatException nfe) {
						Debugger.printThrowable(nfe);
					}
				}
			}
		}
	}

	@Override
	public void stop() {
		final IWorkbench workbench = PlatformUI.getWorkbench();
		if (workbench == null)
			return;
		final Display display = workbench.getDisplay();
		display.syncExec(new Runnable() {
			public void run() {
				if (!display.isDisposed())
					workbench.close();
			}
		});
	}
}
