package org.tekkotsu.ui.storyboard.objects;

import java.util.Comparator;

import org.eclipse.draw2d.Figure;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.LayoutManager;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.jface.resource.ColorRegistry;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PlatformUI;
import org.tekkotsu.ui.storyboard.ResourceRegistry;
import org.tekkotsu.ui.storyboard.model.AbstractModel;
import org.tekkotsu.ui.storyboard.panels.LogBar;
import org.tekkotsu.ui.storyboard.panels.TimeLine;
import org.tekkotsu.ui.storyboard.views.RuntimeView;
import org.tekkotsu.ui.storyboard.views.StoryboardViewer;

public abstract class AbstractViewObject extends Figure {
	public static final String COLOR_BASE = "_color_base";
	public static final String COLOR_FADE = "_color_fade";
	public static final String COLOR_BORDER = "_color_border";

	protected ImageRegistry imgReg;
	protected ColorRegistry colReg;
	protected StoryboardViewer viewer;
	protected TimeLine timeline;
	protected Rectangle constraint;
	private boolean active;

	AbstractViewObject(StoryboardViewer viewer) {
		final Display display = PlatformUI.getWorkbench().getDisplay();
		colReg = new ColorRegistry(display);
		imgReg = ResourceRegistry.getInstance().getImageRegistry();
		this.viewer = viewer;
		this.timeline = viewer.getTimeLine();
	}

	public void forceRepaint() {
		setVisible(false);
		setVisible(true);
	}
	
	public void extendToTime(int value) {
		final Rectangle BOUND = getConstraint();
		final int Y = BOUND.y;
		final int H = BOUND.height;
		
		final int X = timeline.getOffsetOfTime(getModel().getStart());
		final int W = timeline.getOffsetOfTime(value)-X;
		setConstraint(X,Y,W,H);
		forceRepaint();
	}
	
	public boolean getActive() {
		return active;
	}

	public void setActive(boolean b) {
		active = b;
		if (active) {
			getModel().select();
			viewer.viewModel.select(getID());
		} else {
			getModel().deselect();
			viewer.viewModel.deselect(getID());
		}
		updateColors();
	}

	abstract public void updateColors();

	@Override
	public void addNotify() {
		super.addNotify();
		setLayoutManager(getParent().getLayoutManager());
		setConstraint(getConstraint());
	}

	public Rectangle getConstraint() {
		return constraint;
	}

	private void setConstraint(Rectangle constraint) {
		this.constraint = constraint;

		IFigure parent = getParent();
		if (parent != null) {
			LayoutManager layout = parent.getLayoutManager();
			layout.setConstraint(this, constraint);
		}
	}

	public void setConstraint(int x, int y, int w, int h) {
		setConstraint(new Rectangle(x, y, w, h));
	}

	abstract public String getID();

	abstract public AbstractModel getModel();

	public static class BeginComparator implements
			Comparator<AbstractViewObject> {
		@Override
		public int compare(AbstractViewObject a, AbstractViewObject b) {
			int aleft = a.getModel().getRawStart();
			int bleft = b.getModel().getRawStart();

			if( aleft < 0 )
				throw new RuntimeException();
			if( bleft < 0 )
				throw new RuntimeException();
			
			if (aleft < bleft) {
				return -1;
			}

			if (bleft < aleft) {
				return 1;
			}

			//must be consistent with equals()
			return a.getID().compareTo(b.getID());
		}
	}

	public static class EndComparator implements Comparator<AbstractViewObject> {
		@Override
		public int compare(AbstractViewObject a, AbstractViewObject b) {
			AbstractModel am = a.getModel();
			AbstractModel bm = b.getModel();
			
			int aright = am.getRawEnd();
			int bright = bm.getRawEnd();

			if( aright < 0 ) {
				if (am.isFinish()) {
					aright = a.getModel().getEnd();
				} else {
					throw new RuntimeException(am.toString());
				}
			}
			if( bright < 0 ) {
				if( bm.isFinish()) {
					bright = b.getModel().getEnd();
				} else {
					throw new RuntimeException(bm.toString());
				}
			}

			if (aright < bright) {
				return -1;
			}

			if (bright < aright) {
				return 1;
			}

			//must be consistent with equals()
			return a.getID().compareTo(b.getID());
		}
	}

	private AbstractViewObject(int x) {
		constraint = new Rectangle(x, 0, 0, 0);
	}
	
	private static class QueryObject extends AbstractViewObject {
		private AbstractModel model;
		
		QueryObject(int x) {
			super(x);
			
			int start = Math.max(0,x);
			int end = Math.max(0,x);
			
			model = new AbstractModel(start,end,"Query",start,end,true);
		}

		@Override
		public String getID() {
			return "Query";
		}

		@Override
		public AbstractModel getModel() {
			return model;
		}

		@Override
		public void updateColors() {
			//nothing
		}
	}

	public static AbstractViewObject getComparatorQuery(int x) {
		return new QueryObject(x);
	}

	public RuntimeView.TreeObject getRuntimeViewData() {
		return getModel().getRuntimeViewData();
	}
	
	@Override
	public boolean equals(Object o) {
		if( this == o ) return true;
		if( !(o instanceof AbstractViewObject) ) return false;
		AbstractViewObject avo = (AbstractViewObject)o;
		return getModel().equals(avo.getModel());
	}
	
	@Override
	public String toString() {
		return getModel().toString();
	}

	public void setLayoutIndex(int idx) {
        if (getModel().getLayoutIndex() == 0) getModel().setLayoutIndex(idx);
        int xoff = viewer.getTimeLine().getOffsetOfTime(getModel().getStart());
        idx = getModel().getLayoutIndex();
        Rectangle bound = getConstraint();
        if (idx==1) {            
            bound.y=2;
            bound.x=xoff-8;
        } else if (idx==2) {
            bound.y=10;
            bound.x=xoff-12;
        } else if (idx==3) {
            bound.y=10;
            bound.x=xoff-4;
        } else {
            bound.y=LogBar.HEIGHT;
        }
        setConstraint(bound.x,bound.y,bound.width,bound.height); 
	}
}
