package org.tekkotsu.ui.storyboard.views;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.core.runtime.IPath;
import org.eclipse.jface.viewers.Viewer;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.tekkotsu.ui.editor.model.MultiTransitionModel;
import org.tekkotsu.ui.rcp.StoryboardView;
import org.tekkotsu.ui.storyboard.model.AbstractModel;
import org.tekkotsu.ui.storyboard.model.BlockModel;
import org.tekkotsu.ui.storyboard.model.ImageModel;
import org.tekkotsu.ui.storyboard.model.LogModel;
import org.tekkotsu.ui.storyboard.model.TransitionModel;
import org.tekkotsu.ui.util.Debugger;

public class ContentProvider {
	private SortedSet<AbstractModel> blockSetChrono = new TreeSet<AbstractModel>(new ChronologicalComparator());
	
	private List<AbstractModel> blockUnfinished = new ArrayList<AbstractModel>();
	private List<AbstractModel> blockUpdate = new ArrayList<AbstractModel>();

	SAXBuilder builder;

	Element root;

	StoryboardView view;

	int latestTime = 0;

	int offsetTime = -1;

	public ContentProvider(StoryboardView view) {
		builder = new SAXBuilder();
		blockSetChrono.clear();
		blockUnfinished.clear();
		root = new Element("trace");
		this.view = view;
	}

	public void inputChanged(Viewer v, Object oldInput, Object newInput) {
		// nothing
	}

	public void dispose() {
		// nothing
	}

	public Collection<AbstractModel> getElements() {
		return blockSetChrono;
	}

	public List<AbstractModel> getUpdate() {
		List<AbstractModel> updateList = new ArrayList<AbstractModel>();
		updateList.addAll(blockUpdate);
		blockUpdate.clear();
		return updateList;
	}

	public void saveTrace(File file) throws IOException {
		Document document = new Document(root);
		try {
			XMLOutputter outputter = new XMLOutputter(Format.getPrettyFormat());
			Debugger.printDebug(Debugger.DEBUG_ALL, "Attempting to create "
					+ file.getCanonicalPath());
			BufferedOutputStream ostream = new BufferedOutputStream(
					new FileOutputStream(file));
			outputter.output(document, ostream);
			ostream.close();
		} catch (FileNotFoundException e) {
			Debugger.printThrowable(e);
			throw new IOException("Unable to save trace file");
		} catch (IOException e) {
			Debugger.printThrowable(e);
			throw new IOException("Unable to save trace file");
		}
	}

	public void parseNewTrace(IPath path) throws IOException {
		File file = path.toFile();
		
		try {
			Document document = builder.build(file);
			blockSetChrono.clear();
			blockUnfinished.clear();
			offsetTime = -1;
			latestTime = 0;
			parseTrace(document.getRootElement());
		} catch (JDOMException e) {
			Debugger.printThrowable(e);
			throw new IOException("Cannot read input file " + file);
		} catch (IOException e) {
			Debugger.printThrowable(e);
			throw new IOException("Cannot read input file " + file);
		}
	}

	public void getUpdate(String update) {
		try {
			Debugger.printDebug(Debugger.DEBUG_ALL, "Update:" + update);
			StringReader reader = new StringReader(update);
			Document document = builder.build(reader);
			parseTrace(document.getRootElement());
		} catch (Exception e) {
			Debugger.printThrowable(e);
		}
	}

	public void clear() {
		root = new Element("trace");
		blockSetChrono.clear();
		blockUpdate.clear();
		blockUnfinished.clear();
		offsetTime = -1;
		latestTime = 0;
	}

	/**
	 * Mark ending for all unfinished blocks
	 */
	public void endBlocks(int time) {
		//Add new event to the file
		Element event = new Element("event");
		root.addContent(event);

		List<Element> stops = new LinkedList<Element>();
		
		//Collect all unfinished blocks
		for (AbstractModel model : blockUnfinished) {
			if(!model.isFinish()) {
				model.setEnd(AbstractModel.UNKNOWN_TIME,time);

				//stop the state
				Element stop = new Element("statestop");
				stop.setAttribute("id", model.getID());
				stop.setAttribute("time", "" + (model.getEnd() + offsetTime));
			}
		}
		blockUnfinished.clear();

		//Add statestop blocks to event
		for (Element stop : stops) {
			event.addContent(stop);
		}
	}

	/**
	 * @param time
	 *            (raw input from server)
	 * @return time offset relative to the beginning of the trace
	 */
	private int getTime(int time) {
		if (offsetTime == -1) {
			offsetTime = time;
		}

		time = time - offsetTime;
		view.editor.getRobotConnector().sync(time);

		if (time > latestTime) {
			latestTime = time;
		}

		return time;
	}

	private void addBlock(AbstractModel block) {
		blockSetChrono.add(block);
		blockUpdate.add(block);
		if( !block.isFinish() )
		{
			blockUnfinished.add(block);
		}
	}

	private void parseTrace(Element e) {
		final String NAME = e.getName();

		if (NAME.equals("statestart")) {
			parseStateStartTag(e);
		} else if (NAME.equals("statestop")) {
			parseStateStopTag(e);
		} else if (NAME.equals("fire")) {
			parseFireTag(e);
		} else if (NAME.equals("event")) {
			parseEventTag(e);
		} else if (NAME.equals("trace")) {
			parseTraceTag(e);
		}
	}

	private void parseFireTag(Element e) {
		final String ID = e.getAttributeValue("id");
		final String TIME = e.getAttributeValue("time");
		int rawtime = getTime(Integer.parseInt(TIME));
		int protime = rawtime;

		org.tekkotsu.ui.editor.model.AbstractModel trans = view.editor
				.getLayoutData().getPartChild(ID);

		if (trans instanceof MultiTransitionModel) {
			TransitionModel tm = new TransitionModel(ID, rawtime,
					(MultiTransitionModel) trans);
			
			addBlock(tm);
			
			// align the block model ended by this transition
			for (int i = blockUnfinished.size() - 1; i >= 0; i--) {
				Object obj = blockUnfinished.get(i);
				if (obj instanceof BlockModel) {
					BlockModel bm = (BlockModel) obj;
					if (tm.isDestNode(bm.getID())) {
						bm.setEnd(bm.getRawEnd(),protime);
						blockUpdate.add(bm);
						break;
					}
				}
			}
		}
	}

	private void parseStateStopTag(Element e) {
		String id = e.getAttributeValue("id");
		int rawtime = getTime(Integer.parseInt(e.getAttributeValue("time")));
		int protime = rawtime;
		
		// Align stop with the transition that ends this state
		for (int i = blockUnfinished.size() - 1; i >= 0; i--) {
			Object obj = blockUnfinished.get(i);
			if (obj instanceof TransitionModel) {
				TransitionModel tm = (TransitionModel) obj;
				if (tm.isSourceNode(id)) {
					protime = tm.getTime();
					blockUpdate.add(tm);
					break;
				}
			}
		}
		
		//Finish the block
		for (int i = blockUnfinished.size() - 1; i >= 0; i--) {
			Object obj = blockUnfinished.get(i);
			if (obj instanceof BlockModel) {
				BlockModel bm = (BlockModel) obj;
				if (bm.getID().equals(id)) {
					bm.setEnd(rawtime,protime);
					blockUnfinished.remove(bm);
					blockUpdate.add(bm);
					break;
				}
			}
		}
	}

	private void parseStateStartTag(Element e) {
		String id = e.getAttributeValue("id");
		int rawtime = getTime(Integer.parseInt(e.getAttributeValue("time")));
		int protime = rawtime;
		
		// Finish the transition that starts this state
		for (int i = blockUnfinished.size() - 1; i >= 0; i--) {
			Object obj = blockUnfinished.get(i);
			if (obj instanceof TransitionModel) {
				TransitionModel tm = (TransitionModel) obj;
				if (tm.isDestNode(id)) {
					protime = tm.getTime();
					tm.setEnd(rawtime, protime);
					blockUnfinished.remove(tm);
					blockUpdate.add(tm);
					break;
				}
			}
		}
		
		BlockModel bm = new BlockModel(rawtime,id, protime);
		addBlock(bm);
	}

	private void parseEventTag(Element e) {
		root.addContent((Element) e.clone());

		final String TYPE = e.getAttributeValue("type");

		if (TYPE == null) {
			parseChildren(e);
		} else if ("image".equals(TYPE)) {
			parseImageEventTag(e);
		} else if ("userlog".equals(TYPE)) {
			parseUserlogEventTag(e);
		} else if ("log".equals(TYPE)) {
			parseLogEventTag(e);
		}
	}

	private void parseChildren(Element e) {
		@SuppressWarnings("unchecked")
		Iterator<Element> iter = e.getChildren().iterator();

		while (iter.hasNext()) {
			Element child = iter.next();
			parseTrace(child);
		}
	}

	private void parseLogEventTag(Element e) {
		// retrieve system log interface
		String sid = e.getAttributeValue("sid");
		int time = getTime(Integer.parseInt(e.getAttributeValue("time")));
		String content = e.getText();
		String egid = e.getAttributeValue("egid");
		String etid = e.getAttributeValue("etid");
		String icon = e.getAttributeValue("icon");
		String filename = null;
		if( icon != null ) {
			filename = icon;
		} else if( egid != null ) {
			filename = egid;
		}
		
		if (icon != null)
			filename = icon;
		LogModel log = new LogModel(sid, time, filename, etid, egid, content);
		addBlock(log);
	}

	private void parseUserlogEventTag(Element e) {
		// retrieve log interface
		String sid = e.getAttributeValue("sid");
		int time = getTime(Integer.parseInt(e.getAttributeValue("time")));
		String icon = e.getAttributeValue("icon");
		String etid = e.getAttributeValue("etid");
		String egid = e.getAttributeValue("egid");
		String filename = null;
		if( icon != null ) {
			filename = icon;
		} else if( sid != null ) {
			filename = sid;
		}
		String content = e.getText();
		LogModel log = new LogModel(sid, time, filename, etid, egid, content);
		addBlock(log);
	}

	private void parseImageEventTag(Element e) {
		String sid = e.getAttributeValue("sid");
		int time = getTime(Integer.parseInt(e.getAttributeValue("time")));
		String content = e.getChildText("image");
		ImageModel img = new ImageModel(sid, time, content);
		addBlock(img);
	}

	private void parseTraceTag(Element e) {
		root = new Element("trace");
		parseChildren(e);
	}
	
	// (go from right to left)
	private static class ChronologicalComparator implements Comparator<AbstractModel> {
		public int compare(AbstractModel a, AbstractModel b) {
			final int AS = a.getStart();
			final int AE = a.getEnd();
			final int BS = b.getStart();
			final int BE = b.getEnd();
			
			if( AE < 0 && BE > 0 ) return 1;
			if( BE < 0 && AE > 0 ) return -1;
			
			if( AE > BE ) return 1;
			if( AE < BE ) return -1;
			if( AS > BS ) return 1;
			if( AS < BS ) return -1;
			return a.getID().compareTo(b.getID());
		}
	}
}
